package com.famoco.scanenginetestapp.utils;

import android.content.Context;
import android.os.Handler;
import android.util.Log;

import androidx.lifecycle.LiveData;
import androidx.lifecycle.MutableLiveData;

import com.famoco.scanengine.ScanEngine;
import com.famoco.scanengine.ScanEngineListener;
import com.famoco.scanenginetestapp.data.PreferencesManager;

import java.nio.charset.StandardCharsets;
import java.util.Set;

/**
 * Class used to share instance of ScanEngine and centralize configuration/commands/events
 */
public class ScanEngineProvider {
    private static final String TAG = ScanEngineProvider.class.getSimpleName();
    private static final boolean ENABLE_LOG_INFO = true;

    private static ScanEngineProvider sInstance;

    // Data to observe
    private final MutableLiveData<String> mPrivateFirmwareVersion = new MutableLiveData<>();
    public final LiveData<String> firmwareVersion = mPrivateFirmwareVersion;

    private final MutableLiveData<String>  mPrivateBarcodeData = new MutableLiveData<>();
    public final LiveData<String> barcodeData =  mPrivateBarcodeData;

    private MutableLiveData<EngineState>  mPrivateManualScanningState = new MutableLiveData<>();
    public LiveData<EngineState> manualScanningState =  mPrivateManualScanningState;
    private ScanEngine mScanEngine;
    private final ScanEngineListener mScanEngineListener = new ScanEngineListener() {
        @Override
        public void barcodeDetected(byte[] bytes) {
            if (bytes == null) {
                Log.e(TAG, "Barcode detected but no content");
                return;
            }
            String content = new String(bytes, StandardCharsets.UTF_8);
            if (ENABLE_LOG_INFO) {
                Log.i(TAG, "Barcode detected, content: " + content);
            }
            mPrivateBarcodeData.postValue(content);
            int mode = Integer.parseInt(PreferencesManager.getInstance().getTriggerMode());
            if (mode < 2) {
                 mPrivateManualScanningState.postValue(EngineState.READY);
            }
        }

        @Override
        public void onDeviceDetached() {
            if (ENABLE_LOG_INFO) {
                Log.i(TAG, "onDeviceDetached");
            }
            // The scan engine has been detached from the USB slot so we stop the detection.
            // No need to close the communication.
             mPrivateManualScanningState.postValue(EngineState.NONE);
            mPrivateFirmwareVersion.postValue("");
            mScanEngine.stopDetection();
        }

        @Override
        public void onDeviceAttached() {
            if (ENABLE_LOG_INFO) {
                Log.i(TAG, "onDeviceAttached");
            }
            int mode = Integer.parseInt(PreferencesManager.getInstance().getTriggerMode());
            if (mode < 2) {
                 mPrivateManualScanningState.postValue(EngineState.ATTACHED);
            }
            // The ScanEngine has been (re)attached so we can (re)open the communication.
            open();
        }

        @Override
        public void onDeviceReady() {
            if (ENABLE_LOG_INFO) {
                Log.i(TAG, "onDeviceReady");
            }
            // send preference configuration to the attached device
            mPrivateFirmwareVersion.postValue(mScanEngine.getFirmwareVersion());
            configure();
            int mode = Integer.parseInt(PreferencesManager.getInstance().getTriggerMode());
            if (mode < 2) {
                 mPrivateManualScanningState.postValue(EngineState.READY);
            }
        }

        @Override
        public void onDeviceConnectionFail(String s) {
            Log.w(TAG, "onDeviceConnectionFail: " + s);
        }
    };

    private ScanEngineProvider() {
        mPrivateManualScanningState.postValue(EngineState.NONE);
        mPrivateFirmwareVersion.postValue("");
        mPrivateBarcodeData.postValue("");
    }

    public static synchronized ScanEngineProvider getInstance() {
        if (sInstance == null) {
            sInstance = new ScanEngineProvider();
        }
        return sInstance;
    }

    // use application context here
    public void init(Context context) {
        if (mScanEngine != null) {
            Log.w(TAG, "scan engine already initialized");
            return;
        }
        mScanEngine = new ScanEngine(context, mScanEngineListener);
    }

    public void configure() {
        PreferencesManager preferencesManager = PreferencesManager.getInstance();
        String triggerMode = preferencesManager.getTriggerMode();
        switch (triggerMode) {
            case "0":
                setManualTriggerMode(ScanEngine.ManualTriggerMode.NORMAL);
                break;
            case "1":
                setManualTriggerMode(ScanEngine.ManualTriggerMode.CELLPHONE);
                break;
            case "2":
                setPresentationMode(ScanEngine.PresentationMode.NORMAL);
                break;
            case "3":
                setPresentationMode(ScanEngine.PresentationMode.CELLPHONE);
                break;
            case "4":
                setPresentationMode(ScanEngine.PresentationMode.CONTINUE_SCAN);
                break;
        }

        if (mPrivateFirmwareVersion.getValue() != null && !mPrivateFirmwareVersion.getValue().isEmpty()) {
            if (Integer.parseInt(triggerMode) < 2) {
                 mPrivateManualScanningState.postValue(EngineState.READY);
            } else {
                 mPrivateManualScanningState.postValue(EngineState.NONE);
            }
        }

        String rereadDelay = preferencesManager.getRereadDelay();
        boolean isManual = true;
        ScanEngine.ManualTriggerRereadDelay manualDelay =
                ScanEngine.ManualTriggerRereadDelay.NO_DELAY;

        ScanEngine.PresentationModeRereadDelay presentationDelay =
                ScanEngine.PresentationModeRereadDelay.DELAY_500_MS;

        switch (rereadDelay) {
            case "1":
                manualDelay = ScanEngine.ManualTriggerRereadDelay.DELAY_500_MS;
                setManualTriggerModeRereadDelay(ScanEngine.ManualTriggerRereadDelay.DELAY_500_MS);
                break;
            case "2":
                manualDelay = ScanEngine.ManualTriggerRereadDelay.DELAY_2000_MS;
                setManualTriggerModeRereadDelay(ScanEngine.ManualTriggerRereadDelay.DELAY_2000_MS);
                break;
            case "3":
                isManual = false;
                break;
            case "4":
                isManual = false;
                presentationDelay = ScanEngine.PresentationModeRereadDelay.DELAY_750_MS;
                break;
            case "5":
                isManual = false;
                presentationDelay = ScanEngine.PresentationModeRereadDelay.DELAY_1000_MS;
                break;
        }

        if (isManual) {
            setManualTriggerModeRereadDelay(manualDelay);
        } else {
            setPresentationModeRereadDelay(presentationDelay);
        }

        boolean enableAllSymbologies = preferencesManager.getEnableAllSymbologies();
        if (enableAllSymbologies) {
            enableAllSymbologies();
        } else {
            restoreDefaultSymbology();
        }

        Set<String> symbologies = preferencesManager.getEnableSymbologies();
        if (symbologies != null) {
            for (int i = 0; i < 24; i++) {
                String symbology = String.valueOf(i);
                if (symbologies.contains(symbology)) {
                    enableSymbology(symbology);
                }
            }
        }

        boolean supportInverseColorDetection = preferencesManager
                .getSupportInverseColorDetection();
        if (supportInverseColorDetection) {
            enableInverseColor();
        } else {
            disableInverseColor();
        }

        boolean supportInverseColorDetectionOnly = preferencesManager
                .getInverseColorDetectionOnly();
        if (supportInverseColorDetectionOnly) {
            enableInverseColorOnly();
        } else {
            disableInverseColor();
        }

        boolean activate = preferencesManager.getActivateIlluminationLevel();
        if (activate) {
            setIlluminationLevel(ScanEngine.IlluminationLevel.ON);
        } else {
            setIlluminationLevel(ScanEngine.IlluminationLevel.OFF);
        }

        String illuminationLevel = preferencesManager.getIlluminationLevel();
        switch (illuminationLevel) {
            case "0":
                setIlluminationLevel(ScanEngine.IlluminationLevel.LEVEL_1);
                break;
            case "1":
                setIlluminationLevel(ScanEngine.IlluminationLevel.LEVEL_2);
                break;
            case "2":
                setIlluminationLevel(ScanEngine.IlluminationLevel.LEVEL_3);
                break;
            case "3":
                setIlluminationLevel(ScanEngine.IlluminationLevel.LEVEL_4);
                break;
        }

        boolean lock = preferencesManager.getLockConfiguration();
        lockConfiguration(lock);

        boolean display = preferencesManager.getNoDataDisplay();
        setNoDataDisplay(display);

        int timeout = preferencesManager.getDetectionTimeout();
        setDetectionTimeout(timeout);
    }

    private void enableSymbology(String symbology) {
        switch (symbology) {
            case "0":
                enableSymbology(ScanEngine.Symbology.CODE_128);
                break;
            case "1":
                enableSymbology(ScanEngine.Symbology.EAN_8);
                break;
            case "2":
                enableSymbology(ScanEngine.Symbology.EAN_13);
                break;
            case "3":
                enableSymbology(ScanEngine.Symbology.UPC_E0);
                break;
            case "4":
                enableSymbology(ScanEngine.Symbology.UPC_E1);
                break;
            case "5":
                enableSymbology(ScanEngine.Symbology.UPC_A);
                break;
            case "6":
                enableSymbology(ScanEngine.Symbology.INTERLEAVED);
                break;
            case "7":
                enableSymbology(ScanEngine.Symbology.MATRIX);
                break;
            case "8":
                enableSymbology(ScanEngine.Symbology.INDUSTRIAL);
                break;
            case "9":
                enableSymbology(ScanEngine.Symbology.CODE_39);
                break;
            case "10":
                enableSymbology(ScanEngine.Symbology.CODA_BAR);
                break;
            case "11":
                enableSymbology(ScanEngine.Symbology.CODE_93);
                break;
            case "12":
                enableSymbology(ScanEngine.Symbology.GS1_128);
                break;
            case "13":
                enableSymbology(ScanEngine.Symbology.MSI);
                break;
            case "14":
                enableSymbology(ScanEngine.Symbology.CODE_11);
                break;
            case "15":
                enableSymbology(ScanEngine.Symbology.PDF_417);
                break;
            case "16":
                enableSymbology(ScanEngine.Symbology.MICRO_PDF_417);
                break;
            case "17":
                enableSymbology(ScanEngine.Symbology.QR_CODE);
                break;
            case "18":
                enableSymbology(ScanEngine.Symbology.DATA_MATRIX);
                break;
            case "19":
                enableSymbology(ScanEngine.Symbology.MAXI_CODE);
                break;
            case "20":
                enableSymbology(ScanEngine.Symbology.AZTEC);
                break;
            case "21":
                enableSymbology(ScanEngine.Symbology.HANXIN);
                break;
            case "22":
                enableSymbology(ScanEngine.Symbology.CHINA_POSTAL_CODE);
                break;
            case "23":
                enableSymbology(ScanEngine.Symbology.TELEPEN);
                break;
        }
    }

    /**
     * open the communication with the attached device
     */
    public void open() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "open");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        // open scan engine lib => waiting for device detection
        mScanEngine.open();
    }

    /**
     * stop the communication with the detached device
     */
    public void stop() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "stop");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        try {
            // Stop barcode detection
            mScanEngine.stopDetection();
            // close scan engine lib => no more callback received
            mScanEngine.close();
        } catch (Exception e) {
            // nothing to do, we tried to close connection but nothing was running
        }
    }

    /**
     * start barcode detection
     */
    public void startDetection() {
        final PreferencesManager preferencesManager = PreferencesManager.getInstance();
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "startDetection");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        // Start barcode detection
         mPrivateBarcodeData.postValue("");
        mScanEngine.startDetection();
        int mode = Integer.parseInt(preferencesManager.getTriggerMode());
        if (mode < 2) {
             mPrivateManualScanningState.postValue(EngineState.SCANNING);
        }
        new Handler().postDelayed(new Runnable() {
            @Override
            public void run() {
                int mode = Integer.parseInt(preferencesManager.getTriggerMode());
                if ( mPrivateManualScanningState.getValue() == EngineState.SCANNING && mode < 2) {
                     mPrivateManualScanningState.postValue(EngineState.READY);
                }
            }
        }, preferencesManager.getDetectionTimeout());
    }

    /**
     * stop barcode detection
     */
    public void stopDetection() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "stopDetection");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        int mode = Integer.parseInt(PreferencesManager.getInstance().getTriggerMode());
        if (mode < 2) {
             mPrivateManualScanningState.postValue(EngineState.READY);
        }
        // Stop barcode detection
        mScanEngine.stopDetection();
    }

    public void setManualTriggerMode(ScanEngine.ManualTriggerMode triggerMode) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setManualTriggerMode: " + triggerMode);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.setManualTriggerMode(triggerMode);
    }

    public void setPresentationMode(ScanEngine.PresentationMode triggerMode) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setPresentationMode: " + triggerMode);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.setPresentationMode(triggerMode);
    }

    public void setManualTriggerModeRereadDelay(ScanEngine.ManualTriggerRereadDelay delay) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setManualTriggerModeRereadDelay: " + delay);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.setManualTriggerModeRereadDelay(delay);
    }

    public void setPresentationModeRereadDelay(ScanEngine.PresentationModeRereadDelay delay) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setPresentationModeRereadDelay: " + delay);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.setPresentationModeRereadDelay(delay);
    }

    public void enableAllSymbologies() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "enableAllSymbologies");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.enableAllSymbologies();
    }

    public void disableAllSymbologies() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "disableAllSymbologies");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.disableAllSymbologies();
    }

    public void restoreDefaultSymbology() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "restoreDefaultSymbology");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.restoreDefaultSymbology();
    }

    public void enableSymbology(ScanEngine.Symbology symbology) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "enableSymbology: " + symbology);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.enableSymbology(symbology);
    }

    public void disableSymbology(ScanEngine.Symbology symbology) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "disableSymbology: " + symbology);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.disableSymbology(symbology);
    }

    public void enableInverseColor() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "enableInverseColor");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.enableInverseColor();
    }

    public void disableInverseColor() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "disableInverseColor");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.disableInverseColor();
    }

    public void enableInverseColorOnly() {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "enableInverseColorOnly");
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.enableInverseColorOnly();
    }

    public void setIlluminationLevel(ScanEngine.IlluminationLevel level) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setIlluminationLevel: " + level);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.setIlluminationLevel(level);
    }

    public void lockConfiguration(Boolean lock) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "lockConfiguration: " + lock);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.lockConfiguration(lock);
    }

    public void setNoDataDisplay(Boolean display) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setNoDataDisplay: " + display);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        if (display) {
            mScanEngine.enableNoReadMessage();
        } else {
            mScanEngine.disableNoReadMessage();
        }
    }

    public void setDetectionTimeout(int timeout) {
        if (ENABLE_LOG_INFO) {
            Log.i(TAG, "setDetectionTimeout: " + timeout);
        }
        if (mScanEngine == null) {
            Log.w(TAG, "scan engine not initialized");
            return;
        }
        mScanEngine.setDetectionTimeout(timeout);
    }

    public enum EngineState {
        NONE,
        ATTACHED,
        READY,
        SCANNING
    }
}
